/*      overx.c - Merge lines with overstrikes in text files         */

/*      NOTE -- Please read full description at 'mkietf_synopsis[]'   */
/*              later in this program source file.                   */
/*      NOTE -- This program uses strictly ISO 9899:1990 (ISO C)     */
/*              compliant header files, functions, and statements.   */
/*              However, this program declares BOTH strictly ISO C   */
/*              compliant fully specified function prototypes AND    */
/*              traditional C function declarations for greater      */
/*              portability to older systems (eg, SunOS).            */
/*      NOTE -- This program compiles warning free in strict ISO C.  */

/*      Standard Header Files                                        */
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <string.h>
#include <ctype.h>

/*      Standard Logicals for C (Pascal Style)                       */
/*      NOTE -- The specific intent behind the use of the 'Standard  */
/*              Logicals for C' in a program is to ensure:  1) the   */
/*              appearance of an equal sign ('=') ALWAYS indicates   */
/*              an assignment operation; 2) the appearance of an     */
/*              ampersand ('&') ALWAYS indicates EITHER a bitwise    */
/*              AND or a Reference operation; and 3) the appearance  */
/*              of a vertical bar ('|') ALWAYS indicates a bitwise   */
/*              OR operation.                                        */
/*      NOTE -- For strict compliance with ISO 9899 AM1:1994 (ISO C) */
/*              the following logicals should be in lowercase,       */
/*              as per the header file 'iso646.h'.                   */
#ifdef __STDC__
#undef  LT
#undef  GT
#undef  LE
#undef  GE
#undef  EQ
#undef  NE
#undef  AND
#undef  OR
#undef  NOT
#undef  TRUE
#undef  FALSE
#endif
#define LT      <
#define GT      >
#define LE      <=
#define GE      >=
#define EQ      ==
#define NE      !=
#define AND     &&
#define OR      ||
#define NOT     !
#define TRUE    1
#define FALSE   0

/*      Standard Simple Types for C                                  */
/*      NOTE -- The specific intent behind the use of the 'Standard  */
/*              Simple Types for C' is to ensure:  1) the maximal    */
/*              clarity of intended program behavior; and 2) the use */
/*              of a single token for any simple type declaration.   */
/*      NOTE -- For strict compliance with the 'Clean C' dialect,    */
/*              compatible with both ISO C and ISO C++ (draft) and   */
/*              described in 'C: A Reference Manual' by Harbison     */
/*              and Steele (4th Edition, January 1995), the simple   */
/*              type 'boolean' should be respelled 'bool'.           */
#ifdef __STDC__
#undef  uchar
#undef  ushort
#undef  uint
#undef  ulong
#undef  boolean
#endif
#define uchar   unsigned    char
#define ushort  unsigned    short
#define uint    unsigned    int
#define ulong   unsigned    long
#define boolean unsigned    char

#define OVERX_READ_MAX      2048        /* Buffer Read Max Length    */
#define OVERX_FRAG_MAX      50          /* Fragment Max              */

/*********************************************************************/
/*      Global Variable Declarations                                 */
/*********************************************************************/
char    *mkietf_synopsis[] =
    {                                   /* mkietf_synopsis'          */
    "usage: mkietf [-v] filename ...",
    "",
    "                   filename ::= < file_root.file_ext >",
    "              -v   Verbose log output mode",
    "              -p   Write number of lines on each output page",
    "",
    "       Converts contents of each specified file to ietf ",
    "         format, converts characters, removes 1st 7 columns,",
    "         merges overstrike lines. ",
    "       Writes converted file to file name with '-o' suffix  ",
    "         e.g. if filename is 'foo.txt', output is 'foo-o.txt",
    "       Writes to stdout conversion information plus either  ",
    "          'NO ERRORS' or 'ERRORS' with information.         ",
    "       All errors are marked in the output with '#@'.       ",
    "       Errors checked for:  illegal characters,             ",
    "           lines > 72 columns, pages > 58 lines,",
    "           nonblanks except line numbers in first n columns, ",
    "           where n is determined on the first nonempty line. ",
    "",
    "Ex:    mkietf myfile.txt",
    "       (Converts 'myfile.txt' to file 'myfile-o.log'.",
    "        Sends status to stdout. )",
    "",
    NULL
    };                                  /* 'mkietf_synopsis'          */

char    overx_ibuffer[OVERX_READ_MAX+2];/* Input File Buffer         */
char    overx_obuffer[OVERX_READ_MAX+2];/* Output File Buffer        */
char    finalOBuf[OVERX_READ_MAX+2];    /* Final Output File Buffer  */

typedef struct  frag
    {
    char    *ptr;                       /* Fragment string pointer   */
    int     len;                        /* Fragment string length    */
    } FRAG;

FRAG    overx_frag[OVERX_FRAG_MAX];     /* Fragment info table       */
boolean plFlag = FALSE;                 /* page length flag          */
/*********************************************************************/
/*      Global Function Declarations                                 */
/*********************************************************************/
#ifdef __STDC__
extern  int     main (                  /* Mainline                  */
    int     argc,                       /* Argument Count            */
    char    *argv[]);                   /* Argument Vector           */
#else
extern  int     main ();                /* Mainline                  */
#endif

#ifdef __STDC__
static  void    overx_show_synopsis (   /* Display Program Synopsis  */
    void);                              /* No Arguments              */
#else
static  void    overx_show_synopsis (); /* Display Program Synopsis  */
#endif

#ifdef __STDC__
static  void    overx_scan_file (       /* Scan File                 */
    char    *ifn,                       /* Input File Name           */
    boolean vf);                        /* Verbose Flag              */
#else
static  void    overx_scan_file ();     /* Scan File                 */
#endif

#ifdef __STDC__
static  void    overx_error_exit (      /* Log Error and Exit        */
    char    *es);                       /* Error Message String      */
#else
static  void    overx_error_exit ();    /* Log Error and Exit        */
#endif

/*********************************************************************/
/*      Function:       main()                                       */
/*********************************************************************/

#ifdef __STDC__
extern  int     main (                  /* Mainline                  */
    int     argc,                       /* Argument Count            */
    char    *argv[])                    /* Argument Vector           */
#else
extern  int     main (                  /* Mainline                  */
    argc,                               /* Argument Count            */
    argv)                               /* Argument Vector           */
    int     argc;                       /* Argument Count            */
    char    *argv[];                    /* Argument Vector           */
#endif
    {
    int     ai;                         /* Argument Index            */
    char    *ap;                        /* Argument Pointer          */
    boolean vf = FALSE;                 /* Verbose Flag              */

    /*  Check for missing input arguments                            */
    if (argc LT 2)
        {
        overx_show_synopsis ();
        return (0);
        }

    /*  Check for options                                            */
    for (ai = 1; ai LT argc; ai++)
        {
        /*  Get pointer to next input argument                       */
        ap = argv[ai];

        /*  Check for for leading dash (option token)                */
        if (*ap NE '-')
            break;

        /*  Parse zero or more options (case insensitive)            */
        for (ap++; *ap; ap++)
            {
            switch (*ap)
                {
                case 'v':               /* Verbose                   */
                case 'V':               /* Verbose                   */
                    vf = TRUE;
                    break;
                case 'p':               /* put page length in file  */
                    plFlag = TRUE;
                    break;
                default:
                    break;
                }
            }
        }

    /*  Check for missing input filename(s)                          */
    if (ai GE argc)
        {
        overx_show_synopsis ();
        return (0);
        }

    /*  Process all input files                                      */
    for (; ai LT argc; ai++)
        {
        /*  Get next input filename                                  */
        ap = argv[ai];
        if (strlen (ap) EQ 0)
            overx_error_exit ("Missing iuput filename");

        /*  Process next input file                                  */
        overx_scan_file (ap, vf);
        }
   /* (void) printf ("mkietf: Normal termination\n"); */
    return (0);
    }                                   /* 'main'                    */

/*********************************************************************/
/*      Function:       overx_show_synopsis()                        */
/*********************************************************************/
#ifdef __STDC__
static  void    overx_show_synopsis (   /* Display Program Synopsis  */
    void)                               /* No Arguments              */
#else
static  void    overx_show_synopsis (   /* Display Program Synopsis  */
    )                                   /* No Arguments              */
#endif
    {                                   /* 'overx_show_synopsis'     */
    int     si;                         /* Synopsis Index            */

    /*  Display program synopsis                                     */
    for (si = 0; mkietf_synopsis[si]; si++)
        (void) printf ("%s\n", mkietf_synopsis[si]);
    return;
    }                                   /* 'over_show_synopsis'     */

/*********************************************************************/
/*      Function:       overx_scan_file(                             */
/*********************************************************************/
#ifdef __STDC__
static  void    overx_scan_file (       /* Scan File                 */
    char    *ifn,                       /* Input File Name           */
    boolean vf)                         /* Verbose Flag              */
#else
static  void    overx_scan_file (       /* Scan File                 */
    ifn,                                /* Input File Name           */
    vf)                                 /* Verbose Flag              */
    char    *ifn;                       /* Input File Name           */
    boolean vf;                         /* Verbose Flag              */
#endif
    {                                   /* 'overx_scan_file'         */
    boolean ff = FALSE;                 /* Current Formfeed Flag     */
    boolean pf = FALSE;                 /* Pending Formfeed Flag     */
    static  FILE    *ifp = NULL;        /* Input File Pointer        */
    char    *ixp;                       /* Input Excess Pointer      */
    int     ixl;                        /* Input Excess Length       */
    char    *ibp;                       /* Input Buffer Pointer      */
    int     ibl;                        /* Input Buffer Length       */
    int     ibc;                        /* Input Buffer Count        */
    long    ifc;                        /* Input Formfeed Count      */
    long    igc;                        /* Input Graphic Count       */
    long    ilc;                        /* Input Line Count          */
    char    *lfp;                       /* Line Fragment Pointer     */
    int     lfl;                        /* Line Fragment Length      */
    int     lfc;                        /* Line Fragment Count       */
    int     mrl;                        /* Max Raw Length            */
    int     mfl;                        /* Max Fragment Length       */
    int     mfc;                        /* Max Fragment Count        */
    static  FILE    *ofp = NULL;        /* Output File Pointer       */
    char    *ofn;                       /* Output File Name          */
    char    ofs[80];                    /* Output File String        */
    char    *obp;                       /* Output Buffer Pointer     */
    int     obl;                        /* Output Buffer Length      */
    int     obc;                        /* Output Buffer Count       */
    int     owl;                        /* Output Write Length       */
    char    es[80];                     /* Error Message String      */
    char    *fbp;                       /* Final Buffer Pointer      */
    char    *tstr;                      /* Temp String               */
    int     i;                          /* loop int                  */
    int     fbl;                        /* Final Buffer Length       */
    int     j;                          /* temp int                  */
    int     lineLen;                    /* line length               */
    int     maxLineLen=0;               /* line length               */
    int     numMaxLines = 0;            /* number of max length lines*/
    int     badLineCnt=0;               /* bad line count            */
    int     marginCnt = 0;              /* count of margin lines     */
    int     marginCols;                 /* number of margin columns  */
    int     inMargin;                 /* set to 1 if nonblank margin */
    char    lastChar = '\0';            /* last char                 */
    int     firstNonBlankCol = -1;      /* first nonblank column     */
    int     maxPageLines = 0;           /* max lines per page        */
    int     badPageCnt = 0;             /* bad page count            */
    int     pageLines =0;               /* lines per page            */
    int     maxPageNum = 0;             /* max page number           */
    int     pageNum = 0;                /* pageNum                   */
    int     numMaxPages = 0;            /* number of max length pages*/
#define MAXPAGE 100
    int     pageLineCnt[MAXPAGE];       /* page line count           */
#define MAXLINELEN 72
#define MAXPAGELEN 58

    /*  Construct output file name                                   */

    for (j= 0; j<MAXPAGE; j++) { pageLineCnt[j] = 0; }
    ofn = &ofs[0];
    strcpy (ofn, ifn);
    while ((*ofn NE '.') AND (*ofn NE '\0'))
        ofn++;
    strcpy (ofn, "-o.txt");
    ofn = &ofs[0];
    if (strcmp (ifn, ofn) EQ 0)
        {
        (void) sprintf (es,
            "Input filename ('%s') EQ output filename",
            ifn);
        overx_error_exit (es);
        }

    /*  Open input file                                              */
    if (ifp)
        ifp = freopen (ifn, "rb", ifp);
    else
        ifp = fopen (ifn, "rb");
    if (NOT ifp)
        {
        (void) sprintf (es,
            "Error opening input file '%s'",
            ifn);
        overx_error_exit (es);
        }

    /*  Open output file                                             */
    if (ofp)
        ofp = freopen (ofn, "wb", ofp);
    else
        ofp = fopen (ofn, "wb");
    if (NOT ofp)
        {
        (void) sprintf (es,
            "Error opening output file '%s'",
            ofn);
        overx_error_exit (es);
        }

    /*  Read and scan input file                                     */
    ixp = (char *) 0;
    ixl = 0;
    mrl = mfl = mfc = 0;
    for ( ifc = igc = ilc = 0; ;)
        {
        /*   Fill input buffer with binary data (or to end-of-file)  */
        ibp = &overx_ibuffer[0];
        ibp += ixl;
        ibl = OVERX_READ_MAX;
        ibl -= ixl;
        ibl = fread (ibp, 1, ibl, ifp);
        if ((NOT ibl)
        AND (NOT ixl))
            break;
        ibl += ixl;

        /*  Check for pending formfeed (end-of-page)                 */
        if (pf)
            {
            ff = TRUE;
            pf = FALSE;
            }

        /*  Clear the fragment table                                 */
        for (lfc = 0; lfc LT OVERX_FRAG_MAX; lfc++)
            {
            overx_frag[lfc].ptr = (char *) 0;
            overx_frag[lfc].len = 0;
            }

        /*  Scan fragments and convert controls, graphics, and DELs  */
        lfc = 0;
        ibp = &overx_ibuffer[0];
        for (ibc = 0; ibc LT ibl; ibc++, ibp++)
            {
            /*  Save address of next fragment string                 */
            if (NOT overx_frag[lfc].ptr)
                overx_frag[lfc].ptr = lfp = ibp;
            lastChar = *ibp;
            /*  Check for linefeed/newline (end-of-line)             */
            if (*ibp EQ '\n')
                {
                lfl = (int) (ibp - lfp);
                overx_frag[lfc].len = lfl;
                if (lfl GT mfl)
                    mfl = lfl;
                lfc++;
                ilc++;
                break;
                }

            /*  Check for carriage return (end-of-fragment)          */
            if (*ibp EQ '\r')
                {
                lfl = (int) (ibp - lfp);
                overx_frag[lfc].len = lfl;
                if (lfl GT mfl)
                    mfl = lfl;
                lfc++;

                /*  Check for too many fragments                     */
                if (lfc LT OVERX_FRAG_MAX)
                    continue;
                (void) fclose (ofp);
                (void) sprintf (es,
                    "Error too many fragments - line '%ld' in '%s'",
                    ++ilc, ifn);
                overx_error_exit (es);
                }

            /*  Check for formfeed (end-of-page)                     */
            if (*ibp EQ '\f')
                {
                /*  Check for formfeed as FIRST character of line    */
                if (ibp EQ &overx_ibuffer[0])
                    ff = TRUE;
                else
                    pf = TRUE;
                ifc++;
                *ibp = ' ';             /* Is this a good idea??     */
                continue;
                }

            }

        /*  Check for missing linefeed (no end-of-line found)        */
        if (ibc GE ibl)
            {
            /*  Close last line fragment                             */
            lfl = (int) (ibp - lfp);
            overx_frag[lfc].len = lfl;
            if (lfl GT mfl)
                mfl = lfl;
            lfc++;
            ilc++;
            if (lastChar != '\f') (void) printf (
                "Warning no linefeed - line '%ld' in '%s'\n",
                ilc, ifn);
            ibc--;
            ibp--;
            }

        /*  Remember the input buffer excess                         */
        ibc++;
        ibp++;
        ixp = ibp;
        ixl = ibl - ibc;

        /*  Check for maximum raw length or fragments count          */
        if (mrl LT ibc)
            mrl = ibc;
        if (mfc LT lfc)
            mfc = lfc;

        /*  Initialize output buffer to all spaces                   */
        obl = 0;
        obp = &overx_obuffer[0];
        for (obc = 0; obc LT OVERX_READ_MAX; obc++, obp++)
            *obp = ' ';

        /*  Merge fragments to output buffer                         */
        for (lfc = 0; lfc LT OVERX_FRAG_MAX; lfc++)
            {
            /*  Check for last fragement                             */
            lfp = overx_frag[lfc].ptr;
            if (NOT lfp)
                break;

            /*  Check for empty fragment                             */
            lfl = overx_frag[lfc].len;
            if (NOT lfl)
                continue;

            /*  Check for this fragment longer than previous fragment*/
            if (obl LT lfl)
                obl = lfl;

            /*  Check for formfeed needed in this output line        */
            obp = &overx_obuffer[0];

            /*  Merge this fragment                                  */
            for (obc = 0; obc LT lfl; obc++, lfp++, obp++)
                {
                /*  Check for same character in previous fragment    */
                if (*obp EQ *lfp)
                    continue;

                /*  Check for space in previous fragment             */
                if (*obp EQ ' ')
                    {
                    /*  Copy any character over a previous space     */
                    *obp = *lfp;
                    continue;
                    }

                /*  Check for underscore in previous fragment        */
                if (*obp EQ '_')
                    {
                    /*  Check for non-space in current fragment      */
                    if (*lfp NE ' ')
                        *obp = *lfp;
                    continue;
                    }
                }
            }

        /*  Append (carriage return and) linefeed to output buffer   */
        obp = &overx_obuffer[0];
        obp += obl;
#ifdef OSMSDOS
        *obp++ = '\r';
        obl++;
#endif
        *obp++ = '\n';
        obl++;

        /*  Insert formfeed, if necessary, in output buffer          */
        if (ff)
            {
            obp = &overx_obuffer[0];
            obp += obl;
            while (obp NE &overx_obuffer[0])
                {
                *obp = *(obp - 1);
                obp--;
                }
            obp = &overx_obuffer[0];
            *obp = '\f';
            obl++;
            ff = pf = FALSE;
            }

        /*  Write line to output file                                */
        obp = &overx_obuffer[0];
        fbp = &finalOBuf[0];
        fbl = 0;
        for (i=0; i LT obl; i++, obp++) {
            /*  Check for graphic character                          */
            if (*obp == '`') {
                *fbp++ = '\'';
                fbl++;
               }
            else if (*obp == (char) 0xc4) { /* em dash is not converted */
                *fbp++ = '-';
                fbl++;
                }
            else if ((*obp & 0x80) || (*obp EQ 0x7F) ||
                (*obp LE 0x1F && *obp != '\n' && *obp != '\r' &&
                    *obp != '\f'))
                {
                sprintf(fbp,"#@%02x#",((int) *obp) & 0xff);
                fbp += 5;
                fbl += 5;
                igc++;
                }
            else {
                if (*obp == '\f') {
                   pageNum++;
                   if (pageLines < MAXPAGE) pageLineCnt[pageLines]++;
                   if (maxPageLines < pageLines) { 
                      maxPageLines = pageLines;
                      maxPageNum = pageNum;
                      numMaxPages = 1;
                      }
                   else if (maxPageLines == pageLines) numMaxPages++;
                   if (pageLines > MAXPAGELEN) {
                      badPageCnt++;
                      fprintf(ofp,"#@# this page exceeds %d lines: %d\n",MAXPAGELEN,pageLines);
                      }
                   else if (plFlag) {
                      fprintf(ofp,"#@# this page has %d lines\n",pageLines);
                      }
                   pageLines = 0;/* first line on page -- incremented below */
                   }
                *fbp++ = *obp;
                fbl++;
                }

            }
        *fbp = 0;
        fbp--;
        lineLen = (fbl > 0 && *fbp == '\n' ? (fbl > 1 && *(fbp-1) == '\r' ? fbl -2 : fbl -1) 
                                 : (fbl > 1 && *fbp == '\r' ? (*(fbp-1) == '\n' ? fbl -2 : fbl -1)
                                                 : fbl ));      
        inMargin = 0;
       
        if (firstNonBlankCol == -1 && lineLen > 0) {
           for (j = 0, fbp = &finalOBuf[0]; j < lineLen; j++, fbp++ ) {
              if(*fbp != ' ' && (!isdigit(*fbp)) && *fbp != '\f') {
                 firstNonBlankCol = j;
                 break;
                 }
              }
           }
        marginCols =(firstNonBlankCol == -1 || (lineLen > 0 && finalOBuf[0] == '\f')? 0 
                         : lineLen < firstNonBlankCol ? lineLen : firstNonBlankCol); 
        if (marginCols > 0) {
          for (j = marginCols -1, fbp = &finalOBuf[marginCols-1]; j >= 0; j--, fbp--) {
            if (*fbp != ' ' && (!isdigit(*fbp)) && (j > 0 || *fbp != '\f')) {
               fprintf(ofp,"#@# next line has nonblanks in left margin in column: %d\n",j+1);
               marginCnt++;
               inMargin = 1;
               marginCols = 0;
               break;
               }
            }
          lineLen -= marginCols;
          }
       
        if (maxLineLen < lineLen) { 
           maxLineLen = lineLen;
           numMaxLines = 1;
           }
        else if (maxLineLen == lineLen) numMaxLines++;
        if (lineLen > MAXLINELEN) {
           badLineCnt++;
           fprintf(ofp,"#@# next line is > %d characters: %d\n",MAXLINELEN,lineLen);
           }
        pageLines++;
        fbp = &finalOBuf[marginCols];
        fbl -= marginCols;
        owl = fwrite (fbp, 1, fbl, ofp);
        if (owl NE fbl)
            {
            (void) fclose (ofp);
            (void) sprintf (es,
                "Error writing output - line '%ld' file '%s'",
                ilc, ofn);
            overx_error_exit (es);
            }

        /*  Move up remainder of input buffer for next line cycle    */
        if (NOT ixl)
            continue;
        ibp = &overx_ibuffer[0];
        memcpy (ibp, ixp, ixl);
        }

    /*  Close input file                                             */
    (void) fclose (ifp);

    /*  Close output file                                            */
    (void) fclose (ofp);

    /*  Report on this file                                          */
    (void) printf ("\nInput file:      %s\n",ifn);
    (void) printf ("Output file:     %s\n",ofn);
    (void) printf("Cols removed: %4d\n",firstNonBlankCol);
    (void) printf("Total lines:  %4d\n",ilc);
    (void) printf("Longest line: %4d columns on %d lines\n",maxLineLen,numMaxLines);
    (void) printf("Total pages:  %4d\n",ifc);
    (void) printf("Longest page: %4d lines on %d pages\n",maxPageLines,numMaxPages);
    for (j = maxPageLines -1; j >= 0; j--) {
       if (pageLineCnt[j] != 0) {
        (void) printf("Other pages:  %4d lines on %d pages\n",j,pageLineCnt[j]);
          }
       }       
    if (igc > 0 || badLineCnt > 0 || marginCnt > 0 || badPageCnt > 0 ) {
       (void) printf("\n*** ERRORS *** (search for '#@' in output file)\n");
       if (igc > 0) (void) printf("Number of bad characters: %d\n",igc);
       if (badLineCnt > 0) {
          (void) printf("Number of lines > %d columns: %d\n",MAXLINELEN,badLineCnt);
          }
       if (badPageCnt > 0) {
          (void) printf("Number of pages > %d lines: %d\n",MAXPAGELEN,badPageCnt);
          }
       if (marginCnt > 0) 
          (void) printf("Number of lines with nonblanks in 1st %d columns: %d\n",firstNonBlankCol,marginCnt);
       (void) printf("\n");
       }
    else {
       (void) printf("\nNO ERRORS\n\n");
       }

    /*  Check for verbose report mode                                */
    
    if (NOT vf)
        return;
        (void) printf ("\nFormfeeds=%ld\n",ifc);
        (void) printf (
        "Max_Raw_Len=%d\nMax_Fragment_Len=%d\nMax_Fragments=%d\n",
        mrl, mfl-7, mfc);
    return;
    }                                   /* 'overx_scan_file'         */

/*********************************************************************/
/*      Function:       overx_error_exit                             */
/*********************************************************************/
#ifdef __STDC__
static  void    overx_error_exit (      /* Log Error and Exit        */
    char    *es)                        /* Error Message String      */
#else
static  void    overx_error_exit (      /* Log Error and Exit        */
    es)                                 /* Error Message String      */
    char    *es;                        /* Error Message String      */
#endif
    {                                   /* 'overx_error_exit'        */
    (void) printf ("overx: %s\n", es);
    (void) printf ("overx: Abnormal termination\n");
    exit (1);
    return;
    }                                   /* 'overx_error_exit'        */
